@@ -1,24 +1,49 @@
+# This file is part of the django-environ.
+#
+# Copyright (c) 2021, Serghei Iakovlev <egrep@protonmail.ch>
+# Copyright (c) 2013-2021, Daniele Faraglia <daniele.faraglia@gmail.com>
+#
+# For the full copyright and license information, please view
+# the LICENSE.txt file that was distributed with this source code.
+
 """
 Django-environ allows you to utilize 12factor inspired environment
 variables to configure your Django application.
 """
+
 import ast
 import logging
 import os
 import re
 import sys
+import urllib.parse as urlparselib
 import warnings
+from urllib.parse import (
+    parse_qs,
+    ParseResult,
+    unquote_plus,
+    urlparse,
+    urlunparse,
+)
 
 from .compat import (
-    json, urlparselib, urlparse, urlunparse, ParseResult, parse_qs,
-    unquote_plus, DJANGO_POSTGRES, REDIS_DRIVER, ImproperlyConfigured, basestring)
+    DJANGO_POSTGRES,
+    ImproperlyConfigured,
+    json,
+    PYMEMCACHE_DRIVER,
+    REDIS_DRIVER,
+)
+from .fileaware_mapping import FileAwareMapping
+
+try:
+    from os import PathLike
+except ImportError:  # Python 3.5 support
+    from pathlib import PurePath as PathLike
+
+Openable = (str, PathLike)
 
 logger = logging.getLogger(__name__)
 
-VERSION = '0.4.5.2'
-__author__ = 'joke2k'
-__version__ = tuple(VERSION.split('.'))
-
 
 def _cast(value):
     # Safely evaluate an expression node or a string containing a Python
@@ -26,12 +51,12 @@ def _cast(value):
     # https://docs.python.org/3/library/ast.html#ast.literal_eval
     try:
         return ast.literal_eval(value)
-    except ValueError:
+    except (ValueError, SyntaxError):
         return value
 
 
 def _cast_int(v):
-    # return int if possible
+    """Return int if possible."""
     return int(v) if hasattr(v, 'isdigit') and v.isdigit() else v
 
 
@@ -39,13 +64,13 @@ def _cast_urlstr(v):
     return unquote_plus(v) if isinstance(v, str) else v
 
 
-class NoValue(object):
+class NoValue:
 
     def __repr__(self):
-        return '<{0}>'.format(self.__class__.__name__)
+        return '<{}>'.format(self.__class__.__name__)
 
 
-class Env(object):
+class Env:
     """Provide scheme-based lookups of environment variables so that each
     caller doesn't have to pass in `cast` and `default` parameters.
 
@@ -60,6 +85,10 @@ class Env(object):
     NOTSET = NoValue()
     BOOLEAN_TRUE_STRINGS = ('true', 'on', 'ok', 'y', 'yes', '1')
     URL_CLASS = ParseResult
+
+    POSTGRES_FAMILY = ['postgres', 'postgresql', 'psql', 'pgsql', 'postgis']
+    ELASTICSEARCH_FAMILY = ['elasticsearch' + x for x in ['', '2', '5', '7']]
+
     DEFAULT_DATABASE_ENV = 'DATABASE_URL'
     DB_SCHEMES = {
         'postgres': DJANGO_POSTGRES,
@@ -79,7 +108,12 @@ class Env(object):
         'sqlite': 'django.db.backends.sqlite3',
         'ldap': 'ldapdb.backends.ldap',
     }
-    _DB_BASE_OPTIONS = ['CONN_MAX_AGE', 'ATOMIC_REQUESTS', 'AUTOCOMMIT']
+    _DB_BASE_OPTIONS = [
+        'CONN_MAX_AGE',
+        'ATOMIC_REQUESTS',
+        'AUTOCOMMIT',
+        'DISABLE_SERVER_SIDE_CURSORS',
+    ]
 
     DEFAULT_CACHE_ENV = 'CACHE_URL'
     CACHE_SCHEMES = {
@@ -88,11 +122,19 @@ class Env(object):
         'filecache': 'django.core.cache.backends.filebased.FileBasedCache',
         'locmemcache': 'django.core.cache.backends.locmem.LocMemCache',
         'memcache': 'django.core.cache.backends.memcached.MemcachedCache',
-        'pymemcache': 'django.core.cache.backends.memcached.PyLibMCCache',
+        'pymemcache': PYMEMCACHE_DRIVER,
+        'pylibmc': 'django.core.cache.backends.memcached.PyLibMCCache',
         'rediscache': REDIS_DRIVER,
         'redis': REDIS_DRIVER,
+        'rediss': REDIS_DRIVER,
     }
-    _CACHE_BASE_OPTIONS = ['TIMEOUT', 'KEY_PREFIX', 'VERSION', 'KEY_FUNCTION', 'BINARY']
+    _CACHE_BASE_OPTIONS = [
+        'TIMEOUT',
+        'KEY_PREFIX',
+        'VERSION',
+        'KEY_FUNCTION',
+        'BINARY',
+    ]
     _CACHE_BEHAVIOR_PREFIX = 'BEHAVIOR:'
 
     DEFAULT_EMAIL_ENV = 'EMAIL_URL'
@@ -110,19 +152,33 @@ class Env(object):
 
     DEFAULT_SEARCH_ENV = 'SEARCH_URL'
     SEARCH_SCHEMES = {
-        "elasticsearch": "haystack.backends.elasticsearch_backend.ElasticsearchSearchEngine",
-        "elasticsearch2": "haystack.backends.elasticsearch2_backend.Elasticsearch2SearchEngine",
+        "elasticsearch": "haystack.backends.elasticsearch_backend."
+                         "ElasticsearchSearchEngine",
+        "elasticsearch2": "haystack.backends.elasticsearch2_backend."
+                          "Elasticsearch2SearchEngine",
+        "elasticsearch5": "haystack.backends.elasticsearch5_backend."
+                          "Elasticsearch5SearchEngine",
+        "elasticsearch7": "haystack.backends.elasticsearch7_backend."
+                          "Elasticsearch7SearchEngine",
         "solr": "haystack.backends.solr_backend.SolrEngine",
         "whoosh": "haystack.backends.whoosh_backend.WhooshEngine",
         "xapian": "haystack.backends.xapian_backend.XapianEngine",
         "simple": "haystack.backends.simple_backend.SimpleEngine",
     }
+    CLOUDSQL = 'cloudsql'
 
     def __init__(self, **scheme):
+        self.smart_cast = True
+        self.escape_proxy = False
         self.scheme = scheme
 
     def __call__(self, var, cast=None, default=NOTSET, parse_default=False):
-        return self.get_value(var, cast=cast, default=default, parse_default=parse_default)
+        return self.get_value(
+            var,
+            cast=cast,
+            default=default,
+            parse_default=parse_default
+        )
 
     def __contains__(self, var):
         return var in self.ENVIRON
@@ -133,9 +189,9 @@ class Env(object):
         """
         :rtype: str
         """
-        value = self.get_value(var, default=default)
+        value = self.get_value(var, cast=str, default=default)
         if multiline:
-            return value.replace('\\n', '\n')
+            return re.sub(r'(\\r)?\\n', r'\n', value)
         return value
 
     def unicode(self, var, default=NOTSET):
@@ -148,7 +204,10 @@ class Env(object):
         """
         :rtype: bytes
         """
-        return self.get_value(var, cast=str).encode(encoding)
+        value = self.get_value(var, cast=str, default=default)
+        if hasattr(value, 'encode'):
+            return value.encode(encoding)
+        return value
 
     def bool(self, var, default=NOTSET):
         """
@@ -178,13 +237,21 @@ class Env(object):
         """
         :rtype: list
         """
-        return self.get_value(var, cast=list if not cast else [cast], default=default)
+        return self.get_value(
+            var,
+            cast=list if not cast else [cast],
+            default=default
+        )
 
     def tuple(self, var, cast=None, default=NOTSET):
         """
         :rtype: tuple
         """
-        return self.get_value(var, cast=tuple if not cast else (cast,), default=default)
+        return self.get_value(
+            var,
+            cast=tuple if not cast else (cast,),
+            default=default
+        )
 
     def dict(self, var, cast=dict, default=NOTSET):
         """
@@ -196,32 +263,52 @@ class Env(object):
         """
         :rtype: urlparse.ParseResult
         """
-        return self.get_value(var, cast=urlparse, default=default, parse_default=True)
+        return self.get_value(
+            var,
+            cast=urlparse,
+            default=default,
+            parse_default=True
+        )
 
     def db_url(self, var=DEFAULT_DATABASE_ENV, default=NOTSET, engine=None):
         """Returns a config dictionary, defaulting to DATABASE_URL.
 
+        The db method is an alias for db_url.
+
         :rtype: dict
         """
-        return self.db_url_config(self.get_value(var, default=default), engine=engine)
+        return self.db_url_config(
+            self.get_value(var, default=default),
+            engine=engine
+        )
 
     db = db_url
 
     def cache_url(self, var=DEFAULT_CACHE_ENV, default=NOTSET, backend=None):
         """Returns a config dictionary, defaulting to CACHE_URL.
 
+        The cache method is an alias for cache_url.
+
         :rtype: dict
         """
-        return self.cache_url_config(self.url(var, default=default), backend=backend)
+        return self.cache_url_config(
+            self.url(var, default=default),
+            backend=backend
+        )
 
     cache = cache_url
 
     def email_url(self, var=DEFAULT_EMAIL_ENV, default=NOTSET, backend=None):
         """Returns a config dictionary, defaulting to EMAIL_URL.
 
+        The email method is an alias for email_url.
+
         :rtype: dict
         """
-        return self.email_url_config(self.url(var, default=default), backend=backend)
+        return self.email_url_config(
+            self.url(var, default=default),
+            backend=backend
+        )
 
     email = email_url
 
@@ -230,7 +317,10 @@ class Env(object):
 
         :rtype: dict
         """
-        return self.search_url_config(self.url(var, default=default), engine=engine)
+        return self.search_url_config(
+            self.url(var, default=default),
+            engine=engine
+        )
 
     def path(self, var, default=NOTSET, **kwargs):
         """
@@ -249,7 +339,7 @@ class Env(object):
         :returns: Value from environment or default (if set)
         """
 
-        logger.debug("get '{0}' casted as '{1}' with default '{2}'".format(
+        logger.debug("get '{}' casted as '{}' with default '{}'".format(
             var, cast, default
         ))
 
@@ -278,18 +368,28 @@ class Env(object):
             value = self.ENVIRON[var]
         except KeyError:
             if default is self.NOTSET:
-                error_msg = "Set the {0} environment variable".format(var)
+                error_msg = "Set the {} environment variable".format(var)
                 raise ImproperlyConfigured(error_msg)
 
             value = default
 
         # Resolve any proxied values
-        if hasattr(value, 'startswith') and value.startswith('$'):
-            value = value.lstrip('$')
+        prefix = b'$' if isinstance(value, bytes) else '$'
+        escape = rb'\$' if isinstance(value, bytes) else r'\$'
+        if hasattr(value, 'startswith') and value.startswith(prefix):
+            value = value.lstrip(prefix)
             value = self.get_value(value, cast=cast, default=default)
 
-        if cast is None and default is not None and not isinstance(default, NoValue):
-            cast = type(default)
+        if self.escape_proxy and hasattr(value, 'replace'):
+            value = value.replace(escape, prefix)
+
+        # Smart casting
+        if self.smart_cast:
+            if cast is None and default is not None and \
+                    not isinstance(default, NoValue):
+                cast = type(default)
+
+        value = None if default is None and value == '' else value
 
         if value != default or (parse_default and value):
             value = self.parse_value(value, cast)
@@ -326,7 +426,10 @@ class Env(object):
             value = dict(map(
                 lambda kv: (
                     key_cast(kv[0]),
-                    cls.parse_value(kv[1], value_cast_by_key.get(kv[0], value_cast))
+                    cls.parse_value(
+                        kv[1],
+                        value_cast_by_key.get(kv[0], value_cast)
+                    )
                 ),
                 [val.split('=') for val in value.split(';') if val]
             ))
@@ -339,13 +442,14 @@ class Env(object):
             value = tuple([x for x in val if x])
         elif cast is float:
             # clean string
-            float_str = re.sub(r'[^\d,\.]', '', value)
-            # split for avoid thousand separator and different locale comma/dot symbol
-            parts = re.split(r'[,\.]', float_str)
+            float_str = re.sub(r'[^\d,.-]', '', value)
+            # split for avoid thousand separator and different
+            # locale comma/dot symbol
+            parts = re.split(r'[,.]', float_str)
             if len(parts) == 1:
                 float_str = parts[0]
             else:
-                float_str = "{0}.{1}".format(''.join(parts[0:-1]), parts[-1])
+                float_str = "{}.{}".format(''.join(parts[0:-1]), parts[-1])
             value = float(float_str)
         else:
             value = cast(value)
@@ -354,21 +458,14 @@ class Env(object):
     @classmethod
     def db_url_config(cls, url, engine=None):
         """Pulled from DJ-Database-URL, parse an arbitrary Database URL.
-        Support currently exists for PostgreSQL, PostGIS, MySQL, Oracle and SQLite.
 
-        SQLite connects to file based databases. The same URL format is used, omitting the hostname,
-        and using the "file" portion as the filename of the database.
-        This has the effect of four slashes being present for an absolute file path:
-
-        >>> from environ import Env
-        >>> Env.db_url_config('sqlite:////full/path/to/your/file.sqlite')
-        {'ENGINE': 'django.db.backends.sqlite3', 'HOST': '', 'NAME': '/full/path/to/your/file.sqlite', 'PASSWORD': '',
-         'PORT': '', 'USER': ''}
-        >>> Env.db_url_config('postgres://uf07k1i6d8ia0v:wegauwhgeuioweg@ec2-107-21-253-135.compute-1.amazonaws.com:'\
-                              '5431/d8r82722r2kuvn')
-        {'ENGINE': 'django.db.backends.postgresql', 'HOST': 'ec2-107-21-253-135.compute-1.amazonaws.com', 'NAME':
-         'd8r82722r2kuvn', 'PASSWORD': 'wegauwhgeuioweg', 'PORT': 5431, 'USER': 'uf07k1i6d8ia0v'}
+        Support currently exists for PostgreSQL, PostGIS, MySQL, Oracle and
+        SQLite.
 
+        SQLite connects to file based databases. The same URL format is used,
+        omitting the hostname, and using the "file" portion as the filename of
+        the database. This has the effect of four slashes being present for an
+        absolute file path.
         """
         if not isinstance(url, cls.URL_CLASS):
             if url == 'sqlite://:memory:':
@@ -391,13 +488,17 @@ class Env(object):
         if url.scheme == 'sqlite':
             if path == '':
                 # if we are using sqlite and we have no path, then assume we
-                # want an in-memory database (this is the behaviour of  sqlalchemy)
+                # want an in-memory database (this is the behaviour of
+                # sqlalchemy)
                 path = ':memory:'
             if url.netloc:
-                warnings.warn(
-                    'SQLite URL contains host component %r, it will be ignored' % url.netloc, stacklevel=3)
+                warnings.warn('SQLite URL contains host component %r, '
+                              'it will be ignored' % url.netloc, stacklevel=3)
         if url.scheme == 'ldap':
-            path = '{scheme}://{hostname}'.format(scheme=url.scheme, hostname=url.hostname)
+            path = '{scheme}://{hostname}'.format(
+                scheme=url.scheme,
+                hostname=url.hostname,
+            )
             if url.port:
                 path += ':{port}'.format(port=url.port)
 
@@ -410,7 +511,10 @@ class Env(object):
             'PORT': _cast_int(url.port) or '',
         })
 
-        if url.scheme == 'postgres' and path.startswith('/'):
+        if (
+                url.scheme in cls.POSTGRES_FAMILY and path.startswith('/')
+                or cls.CLOUDSQL in path and path.startswith('/')
+        ):
             config['HOST'], config['NAME'] = path.rsplit('/', 1)
 
         if url.scheme == 'oracle' and path == '':
@@ -442,7 +546,7 @@ class Env(object):
             config['ENGINE'] = Env.DB_SCHEMES[config['ENGINE']]
 
         if not config.get('ENGINE', False):
-            warnings.warn("Engine not recognized from url: {0}".format(config))
+            warnings.warn("Engine not recognized from url: {}".format(config))
             return {}
 
         return config
@@ -455,7 +559,16 @@ class Env(object):
         :param backend:
         :return:
         """
-        url = urlparse(url) if not isinstance(url, cls.URL_CLASS) else url
+        if not isinstance(url, cls.URL_CLASS):
+            if not url:
+                return {}
+            else:
+                url = urlparse(url)
+
+        if url.scheme not in cls.CACHE_SCHEMES:
+            raise ImproperlyConfigured(
+                'Invalid cache schema {}'.format(url.scheme)
+            )
 
         location = url.netloc.split(',')
         if len(location) == 1:
@@ -472,7 +585,15 @@ class Env(object):
                 'LOCATION': url.netloc + url.path,
             })
 
-        if not location and url.path and url.scheme in ['memcache', 'pymemcache']:
+        # urlparse('pymemcache://127.0.0.1:11211')
+        # => netloc='127.0.0.1:11211', path=''
+        #
+        # urlparse('pymemcache://memcached:11211/?key_prefix=ci')
+        # => netloc='memcached:11211', path='/'
+        #
+        # urlparse('memcache:///tmp/memcached.sock')
+        # => netloc='', path='/tmp/memcached.sock'
+        if not url.netloc and url.scheme in ['memcache', 'pymemcache']:
             config.update({
                 'LOCATION': 'unix:' + url.path,
             })
@@ -481,28 +602,12 @@ class Env(object):
                 scheme = url.scheme.replace('cache', '')
             else:
                 scheme = 'unix'
-            locations = [scheme + '://' + loc + url.path for loc in url.netloc.split(',')]
-            config['LOCATION'] = locations[0] if len(locations) == 1 else locations
-
-        """
-        TODO: old
-        if url.query:
-            config_options = {}
-            config_behaviors = {}
-            for k, v in urllib.parse.parse_qs(url.query).items():
-                if k.upper() in cls._CACHE_BASE_OPTIONS:
-                    config.update({k.upper(): _cast_int(v[0])})
-                elif k.upper().startswith(cls._CACHE_BEHAVIOR_PREFIX):
-                    config_behaviors.update(
-                        {k.upper().replace(cls._CACHE_BEHAVIOR_PREFIX, '').lower(): _cast_int(v[0])}
-                    )
-                else:
-                    config_options.update({k: _cast_int(v[0])})
-
-            if len(config_behaviors.keys()) > 0:
-                config_options.update({'behaviors': config_behaviors})
-            config['OPTIONS'] = config_options
-        """
+            locations = [scheme + '://' + loc + url.path
+                         for loc in url.netloc.split(',')]
+            if len(locations) == 1:
+                config['LOCATION'] = locations[0]
+            else:
+                config['LOCATION'] = locations
 
         if url.query:
             config_options = {}
@@ -582,23 +687,32 @@ class Env(object):
         path = unquote_plus(path.split('?', 2)[0])
 
         if url.scheme not in cls.SEARCH_SCHEMES:
-            raise ImproperlyConfigured('Invalid search schema %s' % url.scheme)
+            raise ImproperlyConfigured(
+                'Invalid search schema %s' % url.scheme
+            )
         config["ENGINE"] = cls.SEARCH_SCHEMES[url.scheme]
 
         # check commons params
-        params = {}
+        params = {}  # type: dict
         if url.query:
             params = parse_qs(url.query)
             if 'EXCLUDED_INDEXES' in params.keys():
-                config['EXCLUDED_INDEXES'] = params['EXCLUDED_INDEXES'][0].split(',')
+                config['EXCLUDED_INDEXES'] \
+                    = params['EXCLUDED_INDEXES'][0].split(',')
             if 'INCLUDE_SPELLING' in params.keys():
-                config['INCLUDE_SPELLING'] = cls.parse_value(params['INCLUDE_SPELLING'][0], bool)
+                config['INCLUDE_SPELLING'] = cls.parse_value(
+                    params['INCLUDE_SPELLING'][0],
+                    bool
+                )
             if 'BATCH_SIZE' in params.keys():
-                config['BATCH_SIZE'] = cls.parse_value(params['BATCH_SIZE'][0], int)
+                config['BATCH_SIZE'] = cls.parse_value(
+                    params['BATCH_SIZE'][0],
+                    int
+                )
 
         if url.scheme == 'simple':
             return config
-        elif url.scheme in ['solr', 'elasticsearch', 'elasticsearch2']:
+        elif url.scheme in ['solr'] + cls.ELASTICSEARCH_FAMILY:
             if 'KWARGS' in params.keys():
                 config['KWARGS'] = params['KWARGS'][0]
 
@@ -607,13 +721,14 @@ class Env(object):
             path = path[:-1]
 
         if url.scheme == 'solr':
-            config['URL'] = urlunparse(('http',) + url[1:2] + (path,) + ('', '', ''))
+            config['URL'] = urlunparse(
+                ('http',) + url[1:2] + (path,) + ('', '', '')
+            )
             if 'TIMEOUT' in params.keys():
                 config['TIMEOUT'] = cls.parse_value(params['TIMEOUT'][0], int)
             return config
 
-        if url.scheme in ['elasticsearch', 'elasticsearch2']:
-
+        if url.scheme in cls.ELASTICSEARCH_FAMILY:
             split = path.rsplit("/", 1)
 
             if len(split) > 1:
@@ -623,7 +738,9 @@ class Env(object):
                 path = ""
                 index = split[0]
 
-            config['URL'] = urlunparse(('http',) + url[1:2] + (path,) + ('', '', ''))
+            config['URL'] = urlunparse(
+                ('http',) + url[1:2] + (path,) + ('', '', '')
+            )
             if 'TIMEOUT' in params.keys():
                 config['TIMEOUT'] = cls.parse_value(params['TIMEOUT'][0], int)
             config['INDEX_NAME'] = index
@@ -635,7 +752,10 @@ class Env(object):
             if 'STORAGE' in params.keys():
                 config['STORAGE'] = params['STORAGE'][0]
             if 'POST_LIMIT' in params.keys():
-                config['POST_LIMIT'] = cls.parse_value(params['POST_LIMIT'][0], int)
+                config['POST_LIMIT'] = cls.parse_value(
+                    params['POST_LIMIT'][0],
+                    int
+                )
         elif url.scheme == 'xapian':
             if 'FLAGS' in params.keys():
                 config['FLAGS'] = params['FLAGS'][0]
@@ -646,37 +766,64 @@ class Env(object):
         return config
 
     @classmethod
-    def read_env(cls, env_file=None, update=False, **overrides):
-        """
-        Read a .env file into os.environ; if update is True it will update current env with env_file (usefull to
-        update environments variables previously set to default with another env_file).
+    def read_env(cls, env_file=None, overwrite=False, **overrides):
+        """Read a .env file into os.environ.
 
-        If not given a path to a dotenv path, does filthy magic stack backtracking
-        to find manage.py and then find the dotenv.
+        If not given a path to a dotenv path, does filthy magic stack
+        backtracking to find the dotenv in the same directory as the file that
+        called read_env.
 
-        http://www.wellfireinteractive.com/blog/easier-12-factor-django/
+        Existing environment variables take precedent and are NOT overwritten
+        by the file content. ``overwrite=True`` will force an overwrite of
+        existing environment variables.
 
-        https://gist.github.com/bennylope/2999704
+        Refs:
+        - https://wellfire.co/learn/easier-12-factor-django
+        - https://gist.github.com/bennylope/2999704
+
+        :param env_file: The path to the `.env` file your application should
+            use. If a path is not provided, `read_env` will attempt to import
+            the Django settings module from the Django project root.
+        :param overwrite: ``overwrite=True`` will force an overwrite of
+            existing environment variables.
+        :param **overrides: Any additional keyword arguments provided directly
+            to read_env will be added to the environment. If the key matches an
+            existing environment variable, the value will be overridden.
         """
         if env_file is None:
             frame = sys._getframe()
-            env_file = os.path.join(os.path.dirname(frame.f_back.f_code.co_filename), '.env')
+            env_file = os.path.join(
+                os.path.dirname(frame.f_back.f_code.co_filename),
+                '.env'
+            )
             if not os.path.exists(env_file):
-                warnings.warn(
+                logger.info(
                     "%s doesn't exist - if you're not configuring your "
                     "environment separately, create one." % env_file)
                 return
 
         try:
-            with open(env_file) if isinstance(env_file, basestring) else env_file as f:
-                content = f.read()
-        except IOError:
-            warnings.warn(
-                "Error reading %s - if you're not configuring your "
+            if isinstance(env_file, Openable):
+                # Python 3.5 support (wrap path with str).
+                with open(str(env_file)) as f:
+                    content = f.read()
+            else:
+                with env_file as f:
+                    content = f.read()
+        except OSError:
+            logger.info(
+                "%s not found - if you're not configuring your "
                 "environment separately, check this." % env_file)
             return
 
-        logger.debug('Read environment variables from: {0}'.format(env_file))
+        logger.debug('Read environment variables from: {}'.format(env_file))
+
+        def _keep_escaped_format_characters(match):
+            """Keep escaped newline/tabs in quoted strings"""
+            escaped_char = match.group(1)
+            if escaped_char in 'rnt':
+                return '\\' + escaped_char
+            return escaped_char
 
         for line in content.splitlines():
             m1 = re.match(r'\A(?:export )?([A-Za-z_0-9]+)=(.*)\Z', line)
@@ -687,38 +834,50 @@ class Env(object):
                     val = m2.group(1)
                 m3 = re.match(r'\A"(.*)"\Z', val)
                 if m3:
-                    val = re.sub(r'\\(.)', r'\1', m3.group(1))
-                cls.ENVIRON.update({key: str(val)}) if update else cls.ENVIRON.setdefault(key, str(val))
+                    val = re.sub(r'\\(.)', _keep_escaped_format_characters,
+                                 m3.group(1))
+                overrides[key] = str(val)
+            elif not line or line.startswith('#'):
+                # ignore warnings for empty line-breaks or comments
+                pass
+            else:
+                logger.warning('Invalid line: %s', line)
+
+        def set_environ(envval):
+            """Return lambda to set environ.
+
+             Use setdefault unless overwrite is specified.
+             """
+            if overwrite:
+                return lambda k, v: envval.update({k: str(v)})
+            return lambda k, v: envval.setdefault(k, str(v))
+
+        setenv = set_environ(cls.ENVIRON)
 
-        # set defaults
         for key, value in overrides.items():
-            cls.ENVIRON.update({key: value}) if update else cls.ENVIRON.setdefault(key, value)
+            setenv(key, value)
 
 
-class Path(object):
-    """Inspired to Django Two-scoops, handling File Paths in Settings.
-
-        >>> from environ import Path
-        >>> root = Path('/home')
-        >>> root, root(), root('dev')
-        (<Path:/home>, '/home', '/home/dev')
-        >>> root == Path('/home')
-        True
-        >>> root in Path('/'), root not in Path('/other/path')
-        (True, True)
-        >>> root('dev', 'not_existing_dir', required=True)
-        Traceback (most recent call last):
-        environ.environ.ImproperlyConfigured: Create required path: /home/not_existing_dir
-        >>> public = root.path('public')
-        >>> public, public.root, public('styles')
-        (<Path:/home/public>, '/home/public', '/home/public/styles')
-        >>> assets, scripts = public.path('assets'), public.path('assets', 'scripts')
-        >>> assets.root, scripts.root
-        ('/home/public/assets', '/home/public/assets/scripts')
-        >>> assets + 'styles', str(assets + 'styles'), ~assets
-        (<Path:/home/public/assets/styles>, '/home/public/assets/styles', <Path:/home/public>)
-
+class FileAwareEnv(Env):
     """
+    First look for environment variables with ``_FILE`` appended. If found,
+    their contents will be read from the file system and used instead.
+
+    Use as a drop-in replacement for the standard ``environ.Env``:
+
+    .. code-block:: python
+
+        python env = environ.FileAwareEnv()
+
+    For example, if a ``SECRET_KEY_FILE`` environment variable was set,
+    ``env("SECRET_KEY")`` would find the related variable, returning the file
+    contents rather than ever looking up a ``SECRET_KEY`` environment variable.
+    """
+    ENVIRON = FileAwareMapping()
+
+
+class Path:
+    """Inspired to Django Two-scoops, handling File Paths in Settings."""
 
     def path(self, *paths, **kwargs):
         """Create new Path based on self.root and provided paths.
@@ -747,7 +906,7 @@ class Path(object):
 
     def __init__(self, start='', *paths, **kwargs):
 
-        super(Path, self).__init__()
+        super().__init__()
 
         if kwargs.get('is_file', False):
             start = os.path.dirname(start)
@@ -769,12 +928,14 @@ class Path(object):
         return not self.__eq__(other)
 
     def __add__(self, other):
-        return Path(self.__root__, other if not isinstance(other, Path) else other.__root__)
+        if not isinstance(other, Path):
+            return Path(self.__root__, other)
+        return Path(self.__root__, other.__root__)
 
     def __sub__(self, other):
         if isinstance(other, int):
             return self.path('../' * other)
-        elif isinstance(other, basestring):
+        elif isinstance(other, str):
             if self.__root__.endswith(other):
                 return Path(self.__root__.rstrip(other))
         raise TypeError(
@@ -794,7 +955,7 @@ class Path(object):
         return item.__root__.startswith(base_path)
 
     def __repr__(self):
-        return "<Path:{0}>".format(self.__root__)
+        return "<Path:{}>".format(self.__root__)
 
     def __str__(self):
         return self.__root__
@@ -819,7 +980,7 @@ class Path(object):
         absolute_path = os.path.abspath(os.path.join(base, *paths))
         if kwargs.get('required', False) and not os.path.exists(absolute_path):
             raise ImproperlyConfigured(
-                "Create required path: {0}".format(absolute_path))
+                "Create required path: {}".format(absolute_path))
         return absolute_path
 
 
